
SetLogFile("brauer-example.log");
GetVersion();

Attach("../magma/brauer.m");

Zx<x> := PolynomialRing(Integers());
L := NumberField(x^6+3);
G, _, psi := AutomorphismGroup(L);


/***************************************************
 *     Computations in local Brauer groups
 ***************************************************/
rec := LocalBrauerGroup(L,3);

// now H^2(G, L^\times) = H^2(G, M)
// with the following module M
M := rec`M; M;
// the action is a right action:
actM := rec`actM;
{ actM(sig)(actM(tau)(m)) eq actM(tau*sig)(m) : sig in G, tau in G, m in Generators(M)};


// construct a cocycle
H2 := CohomologyGroup(rec`C,2);
c := TwoCocycle(rec`C, H2.1);
// this map has values in some internal structure of Magma
// construct the cocycle as map with values in M
cM := map< car<G,G> -> rec`M | x :-> c(x) @@ rec`f1 >;
// test cocycle relation
{ cM(<rho, tau*sig>) + cM(<tau,sig>) eq   (actM(sig)(cM(<rho,tau>))) + cM(<rho*tau,sig>) 
  : sig in G, tau in G, rho in G };

// or as (multiplicative) cocycle with values in L
cL := map< car<G,G> -> L | x :-> c(x) @@ rec`f1 @@ rec`qM >;
// this is only a cocycle up to finite precision
Minimum( { Valuation( 
  cL(<rho, tau*sig>)* cL(<tau,sig>) - ( (psi(sig)(cL(<rho,tau>))) * cL(<rho*tau,sig>) )
, rec`P )   : sig in G, tau in G, rho in G } );


// compute local fundamental class (up to enough precision)
P := rec`P;
LP, iota := Completion(L, P : Precision := 300);
c := LocalFundamentalClassSerre(LP, pAdicField(LP), 30);
// however, Domain of c is not G
Component(Domain(c),1) eq G;

// construct action on LP, via psi
psiL := map< G -> Aut(LP) | g :-> map< LP -> LP | x :-> x @@ iota @ psi(g) @ iota> >;
c := CocycleLFC(LP, pAdicField(LP), 30 : psi := RightToLeftAction(psiL));
c := LeftToRightCocycle(c);
// test cocycle
Minimum( { Valuation( 
  c(<rho, tau*sig>)* c(<tau,sig>) - ( (psiL(sig)(c(<rho,tau>))) * c(<rho*tau,sig>) )
 )   : sig in G, tau in G, rho in G } );

// read the cocycle in H2
m := func< x | c(x) @@ iota @ rec`qM @ rec`f1 >;
IdentifyTwoCocycle(rec`C, m);


// the local fundamental class can also be computed 
// together with LocalBrauerGroup
rec2 := LocalBrauerGroup(L,3 : lfc);


/* Computation of the local Brauer group for a larger example */
L := NumberField(x^12 - 22*x^10 + 120*x^8 - 252*x^6 + 220*x^4 - 72*x^2 + 4);
time G, _, psi := AutomorphismGroup(L);
psi := map< Domain(psi) -> Codomain(psi) | g :-> psi(g^(-1)) >;
time rec := LocalBrauerGroup(L,2);




/***************************************************
 *     Computations in global Brauer groups
 ***************************************************/

L := SplittingField(x^3+9);
G, _, psi := AutomorphismGroup(L);
psi := map< Domain(psi) -> Codomain(psi) | g :-> psi(g^(-1)) >;

// construc cocycle with local conditions
time c := GlobalCocycle(L, [ <2, 1/2>, <3, 1/2> ]);
// Test cocycle with Galois action from left
G, _, psi := AutomorphismGroup(L);
psi := map< Domain(psi) -> Codomain(psi) | g :-> psi(g^(-1)) >;
{c(sig*tau,rho)*c(sig,tau) eq psi(sig)(c(tau,rho)) * c(sig,tau*rho) : sig in G, tau in G, rho in G};

// verify local conditions
GlobalCocycleInvariants(L,c);

// construct a coboundary
a := map< G -> L | x :-> L.1 > ;
b := map< car<G,G> -> L | g :-> ( a(g[2]) @ psi(g[1]) )*a(g[1])/a(g[1]*g[2]) >;
// and compute its invariants
GlobalCocycleInvariants(L,b);

// multiply b and c and compute its invariants
bc := map< car<G,G> -> L | x :-> b(x)*c(x) >;
GlobalCocycleInvariants(L,bc);


// another cocycle
time d := GlobalCocycle(L, [ <2, 1/2>, <5, 1/2>, <3,0> ]);
GlobalCocycleInvariants(L,d);

// and the invariants of the product
e := map<Domain(d)->Codomain(d) | x:-> c(x)*d(x)>;
GlobalCocycleInvariants(L,e);




