/*
 * artin.m
 * (c) 2008-2009 Ruben Debeerst
 *
 * Methods to compute the Artin map
 *
 * Depends On:
 *   RelAlgKTheory.m (Approximation, MakeTotallyPositive, RealSign, Uniformizer)
 *
 *
 * Methods provided:
 * 
 * localNormResidueSymbol
 *     Compute the norm residue symbol (x, N/M) for a local field
 *     extension N/M. 
 *
 * localNormResidueSymbolAsGlobalIdeal
 *     Computes a global ideal a such that its artin symbol (a, L/K)
 *     represents the local norm residue symbol (x, N/M).
 *
 * globalArtinSymbol
 *     Compute the norm residue symbol (x, L/K) for a global field
 *     extension L/K.
 *     
 */


declare verbose artin, 3;
SetVerbose("artin", 0);

// local functions
forward loc_normres_symb, 
        loc_normres_symb_glob_idl, loc_normres_symb_glob_idl_Z,
        glob_artin_symb, glob_artin_symb_Q;
forward approximation_Z, decomposition_group, primitive_crt,
        primitive_crt_seq, sort_factorization, sort_factorization_Z,
        test_congruence;


intrinsic localNormResidueSymbol(x::., N::FldNum, M::FldNum, PM::.) -> GrpElt, FldAb
{ Let N/M be a global abelian extension and x in M^\times and PM an ideal of M,
  such that there is just one prime ideal PN in N above PM.
  Compute the local norm residue symbol (x, N_PN/M_PM) in Gal(N/M).
  The extension N/M can also be given as abelian field. If the map
  psiH: Gal(N/M) -> Aut(N/M) is not given, the automorphism group of N/M is
  computed first.}
    
    local N2, H, psiH, Na, a;
    require (Type(x) eq FldNumElt and Type(PM) eq RngOrdIdl) or
            (Type(x) eq FldRatElt and Type(PM) eq RngInt) :
            "Expected: x in M and PM ideal in M.";
    
    if BaseField(N) eq M then
        N2 := N;
    else
        vprint artin, 1: "Compute relative field";
        N2 := ext<M|Factorization(PolynomialRing(M)!DefiningPolynomial(N))[1,1]>;
    end if;
    vprint artin, 1: "Compute automorphism group";
    H,_, psiH := AutomorphismGroup(N2,M);
    Na := AbelianExtension(N2);
    
    a := loc_normres_symb(M!x, Na, PM, psiH);
    return a, Na;
end intrinsic;

intrinsic localNormResidueSymbol(x::., Na::FldAb, PM::.) -> GrpElt
{ Let N/M be a global abelian extension and x in M^\times and PM an ideal of M,
  such that there is just one prime ideal PN in N above PM.
  Compute the local norm residue symbol (x, N_PN/M_PM) in Gal(N/M).
  The extension N/M can also be given as abelian field. If the map
  psiH: Gal(N/M) -> Aut(N/M) is not given, the automorphism group of N/M is
  computed first.}
    
    local N,M, H, psiH, a;
    require (Type(x) eq FldNumElt and Type(PM) eq RngOrdIdl) or
            (Type(x) eq FldRatElt and Type(PM) eq RngInt) :
            "Expected: x in M and PM ideal in M.";
    
    N := NumberField(Na);
    M := BaseField(Na);
    H,_, psiH := AutomorphismGroup(N,M);
    return loc_normres_symb(x, Na, PM, psiH);
end intrinsic;

intrinsic localNormResidueSymbol(x::., Na::FldAb, PM::., psiH::Map) -> GrpElt
{ Let N/M be a global abelian extension and x in M^\times and PM an ideal of M,
  such that there is just one prime ideal PN in N above PM.
  Compute the local norm residue symbol (x, N_PN/M_PM) in Gal(N/M).
  The extension N/M can also be given as abelian field. If the map
  psiH: Gal(N/M) -> Aut(N/M) is not given, the automorphism group of N/M is
  computed first.}
    
    local N, H, a, F;
    require (Type(x) eq FldNumElt and Type(PM) eq RngOrdIdl) or
            (Type(x) eq FldRatElt and Type(PM) eq RngInt) :
            "Expected: x in M and PM ideal in M.";
    return loc_normres_symb(x, Na, PM, psiH);
end intrinsic;

// [Exact Algo. (alt), sec 2.3, S. 5]
intrinsic localNormResidueSymbolAsGlobalIdeal(alpha::FldNumElt, F::SeqEnum, PK::RngOrdIdl) -> RngOrdIdl
{ Given the factorization F of the Artin conductor of an abelian extension L/K,
  an element alpha in K and an ideal PK of K sucht that there is just one prime
  ideal PL of L above PK.
  Compute an ideal a of K such that the global Artin symbol (a, L/K) is equal to
  the local norm residue symbol (alpha, L_PL/K_PK). }
    
    require alpha in FieldOfFractions(Order(PK)) :
            "The element alpha and the ideal PK must have the same parent!";
    return loc_normres_symb_glob_idl(alpha, F, PK);
end intrinsic;

intrinsic localNormResidueSymbolAsGlobalIdeal(alpha::FldRatElt, F::SeqEnum, PK::RngInt) -> RngOrdIdl
{ Given the factorization F of the Artin conductor of an abelian extension L/K,
  an element alpha in K and an ideal PK of K sucht that there is just one prime
  ideal PL of L above PK.
  Compute an ideal a of K such that the global Artin symbol (a, L/K) is equal to
  the local norm residue symbol (alpha, L_PL/K_PK). }

    require alpha in Rationals() :
            "The element alpha must be a rational number!";
    return loc_normres_symb_glob_idl_Z(alpha, F, PK);
end intrinsic;

// [Exact Algo. (alt), sec 2.3, S. 6]
intrinsic globalArtinSymbol(a::RngOrdFracIdl, psi::Map) -> GrpElt
{ For an abelian extension L/K, an ideal a in K and psi: Gal(L/K) --> Aut(L).
  Compute the Artin symbel (a, L/K) in Gal(L/K). }
    return glob_artin_symb(a, psi);
end intrinsic;

// Wie globalArtinSymbol fuer K=\Q
intrinsic globalArtinSymbol(a::RngInt, psi::Map) -> GrpElt
{ For an abelian extension L/K, an ideal a in K and psi: Gal(L/K) --> Aut(L).
  Compute the Artin symbel (a, L/K) in Gal(L/K). }
    return glob_artin_symb_Q(a, psi);
end intrinsic;





/***************************************************
 *               Local functions
 ***************************************************/
 
/* loc_normres_symb(x::., Na::FldAb, PM::., psiH::Map) -> GrpElt
 * See localNormResidueSymbol.
 */
function loc_normres_symb(x, Na, PM, psiH)
    local H, N, F, a;
    
    H := Domain(psiH);
    if Type(Codomain(psiH)) eq SeqEnum then
        N := Domain(Codomain(psiH)[1]);
    else
        assert Type(Codomain(psiH)) eq PowMapAut;
        N := Domain(Codomain(psiH));
    end if;
    vprint artin, 1: "Factorization of conductor";
    F := Factorization(Conductor(Na));
    vprint artin, 1: "Compute ideal representing local norm residue symbol";
    a := localNormResidueSymbolAsGlobalIdeal(x, F, PM);
    vprint artin, 1: "Compute global artin symbol for ideal";
    a := globalArtinSymbol(a, psiH);
    return a;
end function;

/* loc_normres_symb_glob_idl(alpha::FldNumElt, F::SeqEnum, PK::RngOrdIdl) -> RngOrdIdl
 * See localNormResidueSymbolAsGlobalIdeal.
 */
function loc_normres_symb_glob_idl(alpha, F, PK)
    local PP, P, pi, e, xi, a, OK, K, I;
    // PK ist erster Faktor in dieser Liste (moeglicherweise mit Exponent 0).
    PP := sort_factorization(F, PK);
    
    OK := Order(PK);
    K := FieldOfFractions(OK);
    //K := Parent(alpha);
    
    e := Valuation(alpha, PK);
    
    if #PP gt 1 then
        pi := primitive_crt_seq([P[1] : P in PP]);
        //pi := Uniformizer([P[1] : P in PP], &*([P[1] : P in PP]))[1];
        if Signature(K) ne 0 then
            pi := MakeTotallyPositive(K!pi, PP[1,1]^2* &*([P[1] : P in PP]) );
        end if;
        //print [Valuation(pi, P[1]) : P in PP];
        // Val_PK(pi)=1, sonst Val_P(pi)=0
        // und pi is totally positive
        xi := Approximation([K!(alpha/pi^e)] cat [K!(1/pi^e)  : i in [1..#PP-1] ],
                             [<P[1], Max(1,P[2]) > : P in PP], &*([P[1]^Max(1,P[2]) : P in PP]),
                             RealPlaces(K));
        // jetzt ist xi/pi^e totally positive
        
    else
        pi := PrimitiveElement(PP[1,1]);
        if Signature(K) ne 0 then
            pi := MakeTotallyPositive(K!pi, PP[1,1]^2);
        end if;
        xi := Approximation([K!(alpha/pi^e)], [<PP[1,1],Max(1,PP[1,2])>], PP[1,1]^Max(1,PP[1,2]), RealPlaces(K));
    end if;

    assert xi in OK;
    
    // Bedingungen testen
    // print "signs2: ", RealSign(xi/pi^e);
    if (Signature(K) ne 0) and (not {r gt 0 : r in RealEmbeddings(xi/pi^e)} eq {true}) then
        error "Real Embeddings sind nicht alle positiv!";
    end if;
    assert test_congruence(OK!xi, pi^e/alpha, PK, Max(1,PP[1,2]));
    for P in Remove(PP,1) do
        assert test_congruence(OK!xi, pi^e, P[1], P[2]);
    end for;
    // xi prim zu PK
    assert Valuation(xi, PK) eq 0;
    
    Factors := Factorization(pi*OK);
    a := (OK!xi) * &*([1*OK] cat [P[1]^(-e*P[2]) : P in Factors | not P[1] eq PK]);
    
    // a prim zu PK
    assert OK!1 in a + PK;
    
    return a;
end function;

/* loc_normres_symb_glob_idl_Z(alpha::FldRatElt, F::SeqEnum, PK::RngInt) -> RngOrdIdl
 * See localNormResidueSymbolAsGlobalIdeal; for ideals of Z.
 */
function loc_normres_symb_glob_idl_Z(alpha, F, PK)
    local PP, a, primes, pi, num, den, Q, s1, Qmal, s2;
    
    PP := sort_factorization_Z(F, PK);
    primes := [Generators(P[1])[1] : P in PP];
    
    e := Valuation(alpha, PK);
    pi := primes[1];
    if pi lt 0 then
        pi := -pi;
    end if;
    xi := approximation_Z([Rationals()!(alpha/pi^e)] cat [Rationals()!(1/pi^e)  : i in [1..#PP-1] ],
                          [<P[1], Max(1,P[2]) > : P in PP], &*([P[1]^Max(1,P[2]) : P in PP]));
    assert test_congruence(Integers()!xi, pi^e/alpha, PK, Max(1,PP[1,2]));
    for P in Remove(PP,1) do
        assert test_congruence(Integers()!xi, pi^e, P[1], P[2]);
    end for;
    
    return (Integers()!xi)*Integers();
end function;

/* glob_artin_symb(a::RngOrdFracIdl, psi::Map) -> GrpElt
 * See globalArtinSymbol.
 */
function glob_artin_symb(a, psi)
    local G, L, OK, OL, factors, fac, g, B, p, P, GD, q, h, H, frob, Q, s1;
    
    G := Domain(psi);
    //L := Domain(Codomain(psi));
    if Type(Codomain(psi)) eq SeqEnum then
        L := Domain(Codomain(psi)[1]);
    else
        assert Type(Codomain(psi)) eq PowMapAut;
        L := Domain(Codomain(psi));
    end if;

    OL := RingOfIntegers(L);
    OK := Order(a);
    B := Basis(OL);
    
    h := Id(G);
    vprintf artin, 2: "Factorization of ideal";
    factors := Factorization(a);
    //print factors;
    vprintf artin, 2: ": %o factors\n", #factors;
    for fac in factors do
        vprintf artin, 3: "factor: %o", fac;
        //print "====\nFactor", fac;
        p := fac[1];
        factors := Factorization(ideal<OL | {OL!b : b in Generators(p) } >);
        P := factors[1,1];
        Q, s1 := quo<OL | P>;
        GD := decomposition_group(P,psi);
        q := Index(OK,p);
        //print q;
        for g in GD do
            for b in B do
                //if Valuation(psi(g)(b)-b^q, P) eq 0 then
                if s1(psi(g)(b))-s1(b)^q ne Zero(Q) then
                   continue g;
                end if;
            end for;
            //durchgekommen
            frob := g;
            break;
        end for;
        //print frob;
        assert &and([s1(psi(frob)(b))-s1(b)^q eq Zero(Q) : b in B]);
        //H := [g : g in GD | {Valuation(psi(g)(b)-b^q, P) ge 1 : b in B} eq {true}];
        //print [*#H, [g^fac[2] : g in H ]   *];
        h := h*(frob^fac[2]);
    end for;
    vprintf artin, 2: "\n";
    //print times;
    return h;
end function;

/* glob_artin_symb_Q(a::RngInt, psi::Map) -> GrpElt
 * See globalArtinSymbol; for ideals of Z.
 */
function glob_artin_symb_Q(a, psi)
    local G, L, OL, factors, fac, g, B, p, P, GD, q, h, H, Q, s1;
    G := Domain(psi);
    //L := Domain(Codomain(psi));
    if Type(Codomain(psi)) eq SeqEnum then
        L := Domain(Codomain(psi)[1]);
    else
        assert Type(Codomain(psi)) eq PowMapAut;
        L := Domain(Codomain(psi));
    end if;
    OL := RingOfIntegers(L);
    B := Basis(OL);
    
    h := Id(G);
    vprintf artin, 2: "Factorization of ideal";
    factors := Factorization(a);
    vprintf artin, 2: ": %o factor(s)\n", #factors;
    for fac in factors do
        vprintf artin, 3: "factor: %o", fac;
        p := Generators(fac[1])[1];
        P := Factorization(p*OL)[1,1];
        //GD := DecompositionGroup(P);
        vprint artin, 3: "Compute decomposition group";
        GD := decomposition_group(P,psi);
        q := #ResidueClassField(Integers(), fac[1]);
        Q, s1 := quo<OL | P>;
        vprint artin, 3: "Find Frobenius";
        H := [g : g in GD | {s1(psi(g)(b)) eq s1(b)^q : b in B} eq {true}];
        assert #H eq 1;
        h := h*(H[1]^fac[2]);
    end for;
    vprint artin, 2: "\n";
    return h;
end function;





/***************************************************
 *              Helper functions
 ***************************************************/

/* Test the congruence xi=rest mod P^e during the computation of the local norm residue symbol.
 */
function test_congruence(xi, rest, P, e)
    if Type(P) eq RngInt then
        OK := Integers();
    else
        OK := Order(P);
    end if;
    K := FieldOfFractions(OK);
    Q, s1 := quo< OK | P^e >;
    Qmal, s2 := MultiplicativeGroup(Q);
    if Type(P) eq RngInt then
        num := Numerator(  K!(rest^(-1)));
        den := Denominator(K!(rest^(-1)));
    else
        num, den := numden(K!(rest^(-1)), P);
    end if;
    //print (s2^(-1))(s1(xi)) + (s2^(-1))(s1(num)) - (s2^(-1))(s1(den)) eq Zero(Qmal);
    
    return (s2^(-1))(s1(xi)) + (s2^(-1))(s1(num)) - (s2^(-1))(s1(den)) eq (s2^(-1))(s1(OK!1));
end function;

/* sort_factorization(Factors::SeqEnum, PK::RngOrdIdl) -> SeqEnum
 * Sort the factors of a factorization, s.t. the first prime ideal appearing
 * in the list is PK (possibly with exponent 0).
 */
function sort_factorization(Factors,PK)
    PP := [];
    for P in Factors do
        if P[1] eq PK then
            Insert(~PP, 1, P);
        else
            Append(~PP, P);
        end if;
    end for;
    if not PP[1,1] eq PK then
        Insert(~PP, 1, <PK, 0>);
    end if;
    
    return PP;
end function;

/* sort_factorization_Z(Factors::SeqEnum, PK::RngInt) -> SeqEnum
 * As above for prime ideals PK of Z.
 */
function sort_factorization_Z(Factors, PK)
    PP := [];
    for P in Factors do
        Q := P[1] meet Integers();
        if Q eq PK then
            Insert(~PP, 1, <Q, P[2]>);
        else
            Append(~PP, <Q, P[2]>);
        end if;
    end for;
    if not PP[1,1] eq PK then
        Insert(~PP, 1, <PK, 0>);
    end if;
    
    return PP;
end function;

/* approximation_Z(alpha::SeqEnum, fac::SeqEnum, IdG::RngInt) -> RngCycElt
 * Algorithm as in the RelAlgKTheory package for ideals of Z.
 */
function approximation_Z(alpha, fac, IdG)
    OK := Integers();
    uniformizer := [Generators(f[1])[1] : f in fac];
    pi := &*[ uniformizer[j]^( -Valuation(alpha[j], fac[j][1]) ) : j in [1..#fac] ] / One(OK);
    c := [];
    for j:=1 to #fac do
       Q, s1 := quo< OK | fac[j][1]^fac[j][2] >;
       Qmal, s2 := MultiplicativeGroup(Q);
       num := Numerator(alpha[j]*pi);
       den := Denominator(alpha[j]*pi);
       Append(~c, OK ! s2( (s2^(-1))(s1(den)) - (s2^(-1))(s1(num)) ) ); 
    end for;

    y := ChineseRemainderTheorem(c, [uniformizer[i]^fac[i,2] : i in [1..#fac]]);
    beta := y * pi;
    if beta lt 0 then
        beta := -beta;
    end if;
    return beta;
end function;

/* primitive_crt(P1::RngOrdIdl, P2::RngOrdIdl) -> RngOrdElt
 * Given two prime ideals P1 and P2. Compute a uniformizing element of P1
 * which ist not in P2.
 */
function primitive_crt(P1, P2)
    OL := Order(P1);
    x := PrimitiveElement(P1);
    // Valuation_P1(x)=1
    if not Valuation(x,P2) eq 0 then
        // Bestimme y in P1^2 mit y notin P2
        y := CRT(P1^2,P2, OL!0, OL!1);
        // Valuation_P1(y)=2
        // Valuation_P2(y)=0
        x := x-y;
    end if;
    
    return x;
end function;

/* primitive_crt_seq(P::SeqEnum) -> RngOrdElt
 * Given prime ideals [P1,P2,...,Pn]. Compute a uniformizing element of P1
 * which ist not in P2,...,Pn.
 */
function primitive_crt_seq(P)
    OL := Order(P[1]);
    x := PrimitiveElement(P[1]);
    // Valuation_P1(x)=1
    // x jetzt so um y abaendern, dass Bewertungen an anderen Stellen
    // auch stimmen
    
    // y hat bei P1 Bewertung >1, damit 
    // die Bewertung von x-y weiterhin 1 ist
    X := [P[1]^2];
    R := [OL!0];
    
    for i in [2..#P] do
        if Valuation(x, P[i]) eq 0 then
            // das soll so bleiben
            // y in P[i], also mit Bewertung >0 waehlen
            // dann hat x-y immer noch Bewertung 0
            Append(~X, P[i]);
            Append(~R, OL!0);
            
        else
            // y mit Bewertung 0 waehlen,
            // dann hat x-y auch Bewertung 0
            Append(~X, P[i]);
            Append(~R, OL!1);
        end if;
    end for;
    
    y := CRT(R, X);
    return x-y;
end function;

/* 
 * Let L/K be an extension of global fields with Galois group G
 * and psi: G --> Aut(L/K).
 * Compute the decomposition group of the prime ideal PK
 * as subgroup of G.
 * 
 */
function decomposition_group(PK, psi)
//intrinsic decompGroup(PN::RngOrdIdl, H::Grp, psi::Map) -> Grp
    G := Domain(psi);
    return sub< G | {sigma : sigma in G |
                        &and([ psi(sigma)(b) in PK : b in Generators(PK)])
                    } >;
end function;

